"""Additional utilities"""

import functools
import threading
import time
from collections.abc import Callable
from typing import Any, TypeVar

import pandas as pd
from structlog import get_logger

RT = TypeVar('RT')  # return type


def rm_keys(
    keys: Any | list[Any],
    dct: dict[Any, Any],
) -> dict[Any, Any]:
    """Return a copy with keys removed from dictionary"""
    if not isinstance(keys, list):
        keys = [keys]
    return {k: v for k, v in dct.items() if k not in keys}


def pretty_timedelta(seconds: int) -> str:
    """Converts timedelta in seconds to human-readable string

    Args:
        seconds: time delta in seconds

    Returns:
        timedelta as pretty string
    """
    sign = '-' if seconds < 0 else ''
    seconds = abs(int(seconds))
    days, seconds = divmod(seconds, 86400)
    hours, seconds = divmod(seconds, 3600)
    minutes, seconds = divmod(seconds, 60)
    if days > 0:
        return f'{sign}{days}d{hours}h{minutes}m{seconds}s'
    elif hours > 0:
        return f'{sign}{hours}h{minutes}m{seconds}s'
    elif minutes > 0:
        return f'{sign}{minutes}m{seconds}s'
    else:
        return f'{sign}{seconds}s'


def throttle(calls: int, seconds: int = 1) -> Callable[[Callable[..., RT]], Callable[..., RT]]:
    """Decorator for throttling a function to number of calls per seconds

    Args:
        calls: number of calls per interval
        seconds: number of seconds in interval

    Returns:
        wrapped function
    """
    if not isinstance(calls, int):
        msg = 'number of calls must be integer'
        raise ValueError(msg)
    if not isinstance(seconds, int):
        msg = 'number of seconds must be integer'
        raise ValueError(msg)

    def decorator(func: Callable[..., RT]) -> Callable[..., RT]:
        # keeps track of the last calls
        last_calls: list[float] = []
        lock = threading.Lock()

        @functools.wraps(func)
        def wrapper(*args, **kwargs) -> RT:
            nonlocal last_calls
            with lock:
                curr_time = time.time()
                # Remove old calls
                last_calls = [call for call in last_calls if call > curr_time - seconds]

                if len(last_calls) >= calls:
                    sleep_time = last_calls[0] + seconds - curr_time
                    logger = get_logger()
                    logger.debug('stalling call', func=func.__name__, secs=sleep_time)
                    time.sleep(sleep_time)

                resp = func(*args, **kwargs)
                last_calls.append(time.time())
                return resp

        return wrapper

    return decorator


def implode(df: pd.DataFrame, cols: str | list[str]) -> pd.DataFrame:
    """The inverse of Pandas' explode"""
    if not isinstance(cols, list):
        cols = [cols]
    orig_cols = df.columns
    grp_cols = [col for col in df.columns if col not in cols]
    df = df.groupby(grp_cols, group_keys=True, dropna=False).aggregate({col: lambda x: x.tolist() for col in cols})
    df.reset_index(inplace=True)
    df = df.loc[:, orig_cols]
    return df



"""Some helper functions for HiGHS (https://highs.dev/)

`pyomo` and `highspy` need to be installed, consider `pip install 'pytanis[all]'`.

ToDo:
    * Introduce a function `check_model_vars` that checks the names of variables and sets to be alphanumeric
      before reading in a solution in `set_solution_from_file`.
"""

import re
from collections.abc import Iterator

from pyomo.contrib.appsi.solvers import Highs
from pyomo.core.base.PyomoModel import ConcreteModel


def read_sol_file(file_name: str) -> Iterator[tuple[str, float]]:
    """Read a solution file from HiGHS solver with default output style

    Attention: We assume here that your variable names are alphanumeric!
               No underscores, no dashes, etc.!
    """
    line_re = re.compile(r'(\w+)(?:\((\w+)\))?(_binary_indicator_var)? ([.\w-]+)')

    with open(file_name, encoding='utf8') as fh:
        while True:
            line = fh.readline()
            if line.startswith('# Columns'):
                break
        for line in fh.readlines():
            if line.startswith('#'):
                break
            if (match_obj := line_re.match(line.strip())) is None:
                msg = f'Could not interpret line: {line}'
                raise RuntimeError(msg)
            else:
                var_name, idx, binary, val = match_obj.groups()
            val = float(val)
            binary = binary.replace('_', '.', 1) if binary else ''

            if idx is None:
                yield f'{var_name}{binary}', val
            else:
                idx = idx.replace('_', ',')
                yield f'{var_name}[{idx}]{binary}', val


def set_solution_from_file(model: ConcreteModel, file_name: str):
    """Given a HiGHS solution file set the variables of a Pyomo model accordingly.

    This is a workaround to set a Pyomo model's variables to the solution
    from a HiGHS solution file.
    """
    # just to initialize we read it in using HiGHS. The result is incorrect though,
    # as the order of variables is mixed up quite often. We fix this below!
    opt = Highs()
    opt.set_instance(model)
    opt._solver_model.readSolution(file_name, 0)
    opt._sol = opt._solver_model.getSolution()
    opt.load_vars()

    # read the actual mapping of the variable names to the values
    file_sol = dict(read_sol_file(file_name))

    # overwrite the values of the variables again using the symbolic names from the file
    for v, ref_info in opt._referenced_variables.items():
        using_cons, using_sos, using_obj = ref_info
        if using_cons or using_sos or (using_obj is not None):
            var = opt._vars[v][0]
            var.set_value(file_sol[var.name], skip_validation=True)



"""Return types of the HelpDesk / LiveChat API

Documentation: https://api.helpdesk.com/docs

ToDo:
    * Implement the types below correctly instead of using `Extra.Allow`
    * Find out why `extra=Extra.allow` causes mypy to fail. Seems like a bug in pydantic.
"""

from pydantic import BaseModel


class Id(BaseModel):
    ID: str


class Agent(BaseModel, extra='allow'):  # type: ignore
    pass


class Team(BaseModel, extra='allow'):  # type: ignore
    pass


class Message(BaseModel, extra='allow'):  # type: ignore
    text: str


class Requester(BaseModel):
    email: str
    name: str


class Assignment(BaseModel, extra='allow'):  # type: ignore
    team: Id
    agent: Id


class NewTicket(BaseModel, extra='allow'):  # type: ignore
    """Object that needs to be sent when creating a NEW ticket"""

    message: Message
    requester: Requester
    status: str | None = None  # ToDo: Rather use an Enum instead
    subject: str | None = None
    teamIDs: list[str] | None = None  # noqa: N815
    assignment: Assignment | None = None


class Ticket(BaseModel, extra='allow'):  # type: ignore
    """Actual ticket as returned by the API"""



from importlib.metadata import PackageNotFoundError, version

import structlog.stdlib

from pytanis.config import get_cfg
from pytanis.google import GSheetClient
from pytanis.helpdesk import HelpDeskClient
from pytanis.pretalx import PretalxClient

try:
    __version__ = version('pytanis')
except PackageNotFoundError:  # pragma: no cover
    __version__ = 'unknown'
finally:
    del version, PackageNotFoundError

__all__ = ['__version__', 'GSheetClient', 'PretalxClient', 'HelpDeskClient', 'get_cfg']

# transform structlog into a logging-friendly package
# use `logging.basicConfig(level=logging.INFO, stream=sys.stdout)` as usual
# taken from https://www.structlog.org/en/stable/standard-library.html#rendering-within-structlog
structlog.configure(
    processors=[
        # If log level is too low, abort pipeline and throw away log entry.
        structlog.stdlib.filter_by_level,
        # Add the name of the logger to event dict.
        structlog.stdlib.add_logger_name,
        # Add log level to event dict.
        structlog.stdlib.add_log_level,
        # Perform %-style formatting.
        structlog.stdlib.PositionalArgumentsFormatter(),
        # Add a timestamp in ISO 8601 format.
        structlog.processors.TimeStamper(fmt='iso'),
        # If the "stack_info" key in the event dict is true, remove it and
        # render the current stack trace in the "stack" key.
        structlog.processors.StackInfoRenderer(),
        # If the "exc_info" key in the event dict is either true or a
        # sys.exc_info() tuple, remove "exc_info" and render the exception
        # with traceback into the "exception" key.
        structlog.processors.format_exc_info,
        # If some value is in bytes, decode it to a unicode str.
        structlog.processors.UnicodeDecoder(),
        # Add callsite parameters.
        structlog.processors.CallsiteParameterAdder({
            structlog.processors.CallsiteParameter.FILENAME,
            structlog.processors.CallsiteParameter.FUNC_NAME,
            structlog.processors.CallsiteParameter.LINENO,
        }),
        # Render the final event dict as JSON.
        structlog.processors.JSONRenderer(),
    ],
    # `wrapper_class` is the bound logger that you get back from
    # get_logger(). This one imitates the API of `logging.Logger`.
    wrapper_class=structlog.stdlib.BoundLogger,
    # `logger_factory` is used to create wrapped loggers that are used for
    # OUTPUT. This one returns a `logging.Logger`. The final value (a JSON
    # string) from the final processor (`JSONRenderer`) will be passed to
    # the method of the same name as that you've called on the bound logger.
    logger_factory=structlog.stdlib.LoggerFactory(),
    # Effectively freeze configuration after creating the first bound
    # logger.
    cache_logger_on_first_use=True,
)



"""Return types of the Pretalx API

Documentation: https://docs.pretalx.org/api/resources/index.html

Attention: Quite often the API docs and the actual results of the API differ!

ToDo:
    * Find out why `extra=Extra.allow` causes mypy to fail. Seems like a bug in pydantic.
"""

from datetime import date, datetime
from enum import Enum

from pydantic import BaseModel, Field


class Me(BaseModel):
    name: str
    email: str
    local: str | None = None
    timezone: str


class MultiLingualStr(BaseModel, extra='allow'):  # type: ignore
    # ToDo: Add here more available languages, not mentioned in the API
    en: str | None = None  # we assume though that english is always given to simplify things
    de: str | None = None


class URLs(BaseModel):
    base: str
    schedule: str
    login: str
    feed: str


class Event(BaseModel):
    name: MultiLingualStr
    slug: str
    is_public: bool
    date_from: date
    date_to: date | None = None
    timezone: str
    urls: URLs


class SpeakerAvailability(BaseModel):
    # ToDo: Check the datatypes here again, not mentioned in the API
    id: int  # noqa: A003
    start: str
    end: str
    allDay: str = Field(..., alias='all_day')  # noqa: N815


class AnswerQuestionRef(BaseModel):
    id: int  # noqa: A003
    question: MultiLingualStr


class Option(BaseModel):
    id: int  # noqa: A003
    answer: MultiLingualStr


class Answer(BaseModel):
    id: int  # noqa: A003
    answer: str
    answer_file: str | None = None
    question: AnswerQuestionRef
    submission: str | None = None
    review: int | None = None
    person: str | None = None
    options: list[Option]


class SubmissionSpeaker(BaseModel):
    code: str
    name: str
    biography: str | None = None
    avatar: str | None = None
    email: str | None = None


class Speaker(SubmissionSpeaker):
    submissions: list[str]  # submission codes
    availabilities: list[SpeakerAvailability] | None = None  # maybe needs organizer permissions?
    answers: list[Answer] | None = None  # maybe needs organizer permissions?


class Slot(BaseModel):
    start: datetime
    end: datetime
    room: MultiLingualStr
    room_id: int


class Resource(BaseModel):
    resource: str
    description: str


class State(Enum):
    submitted = 'submitted'
    accepted = 'accepted'
    rejected = 'rejected'  # is "Not accepted" in WebUI
    confirmed = 'confirmed'
    withdrawn = 'withdrawn'
    canceled = 'canceled'
    deleted = 'deleted'


class Submission(BaseModel):
    code: str
    speakers: list[SubmissionSpeaker]
    created: datetime | None = None  # needs organizer permissions
    title: str
    submission_type: MultiLingualStr
    submission_type_id: int
    track: MultiLingualStr | None = None
    track_id: int | None = None
    state: State
    pending_state: State | None = None  # needs organizer permissions
    abstract: str
    description: str
    duration: int | None = None
    do_not_record: bool
    is_featured: bool
    content_locale: str  # e.g. "de", "en"
    slot: Slot | None = None  # only available after schedule_web release
    slot_count: int
    image: str | None = None
    answers: list[Answer] | None = None  # needs organizer permissions and `questions` query parameter
    notes: str | None = None  # needs organizer permissions
    internal_notes: str | None = None  # needs organizer permissions
    resources: list[Resource]
    tags: list[str] | None = None  # needs organizer permissions
    tag_ids: list[int] | None = None  # needs organizer permissions


class Talk(Submission):
    pass


class User(BaseModel):
    name: str
    email: str


class Review(BaseModel):
    id: int  # noqa: A003
    submission: str
    user: str
    text: str | None = None
    score: float | None = None  # converted from str if present
    created: datetime
    updated: datetime
    answers: list[str]  # ToDo: Check this type


class RoomAvailability(BaseModel):
    start: datetime
    end: datetime


class Room(BaseModel):
    id: int  # noqa: A003
    name: MultiLingualStr
    description: MultiLingualStr
    capacity: int | None = None
    position: int | None = None
    speaker_info: MultiLingualStr | None = None
    availabilities: list[RoomAvailability] | None = None  # needs organizer permissions


class QuestionRequirement(Enum):
    optional = 'optional'
    required = 'required'
    after_deadline = 'after deadline'


class Question(BaseModel):
    id: int  # noqa: A003
    variant: str
    target: str
    question: MultiLingualStr
    help_text: MultiLingualStr
    question_required: QuestionRequirement
    deadline: datetime | None = None
    required: bool
    read_only: bool | None = None
    freeze_after: datetime | None = None
    options: list[Option]
    default_answer: str | None = None
    contains_personal_data: bool
    min_length: int | None = None
    max_length: int | None = None
    is_public: bool
    is_visible_to_reviewers: bool


class Tag(BaseModel):
    tag: str
    description: MultiLingualStr
    color: str



"""Client for the Pretalx API

Documentation: https://docs.pretalx.org/api/resources/index.html

ToDo:
    * add additional parameters explicitly like querying according to the API
"""

from collections.abc import Iterator
from typing import Any, TypeAlias, TypeVar, cast

import httpx
from httpx import URL, QueryParams, Response
from httpx_auth import HeaderApiKey
from pydantic import BaseModel
from structlog import get_logger
from tqdm.auto import tqdm

from pytanis.config import Config, get_cfg
from pytanis.pretalx.types import Answer, Event, Me, Question, Review, Room, Speaker, Submission, Tag, Talk
from pytanis.utils import rm_keys, throttle

_logger = get_logger()


T = TypeVar('T', bound=BaseModel)
JSONObj: TypeAlias = dict[str, Any]
"""Type of a JSON object (without recursion)"""
JSONLst: TypeAlias = list[JSONObj]
"""Type of a JSON list of JSON objects"""
JSON: TypeAlias = JSONObj | JSONLst
"""Type of the JSON response as returned by the Pretalx API"""
QueryParamType: TypeAlias = dict[Any, Any | list[Any]] | QueryParams
"""Type for the optional parameters to the Pretalx API"""


class PretalxClient:
    """Client for the Pretalx API"""

    def __init__(self, config: Config | None = None, *, blocking: bool = False):
        if config is None:
            config = get_cfg()
        self._config = config
        self._get_throttled = self._get
        self.blocking = blocking
        self.set_throttling(1, 2)  # we are nice by default

    def set_throttling(self, calls: int, seconds: int):
        """Throttle the number of calls per seconds to the Pretalx API"""
        _logger.info('throttling', calls=calls, seconds=seconds)
        self._get_throttled = throttle(calls, seconds)(self._get)

    def _get(self, endpoint: str, params: QueryParamType | None = None) -> Response:
        """Retrieve data via GET request"""
        auth = HeaderApiKey(self._config.Pretalx.api_token, header_name='Authorization')
        url = URL('https://pretalx.com/').join(endpoint).copy_merge_params(params)
        _logger.info(f'GET: {url}')
        # we set the timeout to 60 seconds as the Pretalx API is quite slow
        return httpx.get(url, auth=auth, timeout=60.0)

    def _get_one(self, endpoint: str, params: QueryParamType | None = None) -> JSON:
        """Retrieve a single resource result"""
        resp = self._get_throttled(endpoint, params)
        resp.raise_for_status()
        return resp.json()

    def _resolve_pagination(self, resp: JSONObj) -> Iterator[JSONObj]:
        """Resolves the pagination and returns an iterator over all results"""
        yield from resp['results']
        while (next_page := resp['next']) is not None:
            endpoint = URL(next_page).path
            resp = cast(JSONObj, self._get_one(endpoint, URL(next_page).params))
            _log_resp(resp)
            yield from resp['results']

    def _get_many(self, endpoint: str, params: QueryParamType | None = None) -> tuple[int, Iterator[JSONObj]]:
        """Retrieves the result count as well as the results as iterator"""
        resp = self._get_one(endpoint, params)
        _log_resp(resp)
        if isinstance(resp, list):
            return len(resp), iter(resp)
        elif self.blocking:
            _logger.debug('blocking resolution of pagination...')
            return resp['count'], iter(list(tqdm(self._resolve_pagination(resp), total=resp['count'])))
        else:
            _logger.debug('non-blocking resolution of pagination...')
            return resp['count'], self._resolve_pagination(resp)

    def _endpoint_lst(
        self,
        type: type[T],  # noqa: A002
        event_slug: str,
        resource: str,
        *,
        params: QueryParamType | None = None,
    ) -> tuple[int, Iterator[T]]:
        """Queries an endpoint returning a list of resources"""
        endpoint = f'/api/events/{event_slug}/{resource}/'
        count, results = self._get_many(endpoint, params)
        t_results = iter(_logger.debug('result', resp=r) or type.parse_obj(r) for r in results)
        return count, t_results

    def _endpoint_id(
        self,
        type: type[T],  # noqa: A002
        event_slug: str,
        resource: str,
        id: int | str,  # noqa: A002
        *,
        params: QueryParamType | None = None,
    ) -> T:
        """Query an endpoint returning a single resource"""
        endpoint = f'/api/events/{event_slug}/{resource}/{id}/'
        result = self._get_one(endpoint, params)
        _logger.debug('result', resp=result)
        return type.model_validate(result)

    def me(self) -> Me:
        """Returns what Pretalx knows about myself"""
        result = self._get_one('/api/me')
        return Me.model_validate(result)

    def event(self, event_slug: str, *, params: QueryParamType | None = None) -> Event:
        """Returns detailed information about a specific event"""
        endpoint = f'/api/events/{event_slug}/'
        result = self._get_one(endpoint, params)
        _logger.debug('result', resp=result)
        return Event.model_validate(result)

    def events(self, *, params: QueryParamType | None = None) -> tuple[int, Iterator[Event]]:
        """Lists all events and their details"""
        count, results = self._get_many('/api/events/', params)
        events = iter(_logger.debug('result', resp=r) or Event.model_validate(r) for r in results)
        return count, events

    def submission(self, event_slug: str, code: str, *, params: QueryParamType | None = None) -> Submission:
        """Returns a specific submission"""
        return self._endpoint_id(Submission, event_slug, 'submissions', code, params=params)

    def submissions(self, event_slug: str, *, params: QueryParamType | None = None) -> tuple[int, Iterator[Submission]]:
        """Lists all submissions and their details"""
        return self._endpoint_lst(Submission, event_slug, 'submissions', params=params)

    def talk(self, event_slug: str, code: str, *, params: QueryParamType | None = None) -> Talk:
        """Returns a specific talk"""
        return self._endpoint_id(Talk, event_slug, 'talks', code, params=params)

    def talks(self, event_slug: str, *, params: QueryParamType | None = None) -> tuple[int, Iterator[Talk]]:
        """Lists all talks and their details"""
        return self._endpoint_lst(Talk, event_slug, 'talks', params=params)

    def speaker(self, event_slug: str, code: str, *, params: QueryParamType | None = None) -> Speaker:
        """Returns a specific speaker"""
        return self._endpoint_id(Speaker, event_slug, 'speakers', code, params=params)

    def speakers(self, event_slug: str, *, params: QueryParamType | None = None) -> tuple[int, Iterator[Speaker]]:
        """Lists all speakers and their details"""
        return self._endpoint_lst(Speaker, event_slug, 'speakers', params=params)

    def review(self, event_slug: str, id: int, *, params: QueryParamType | None = None) -> Review:  # noqa: A002
        """Returns a specific review"""
        return self._endpoint_id(Review, event_slug, 'reviews', id, params=params)

    def reviews(self, event_slug: str, *, params: QueryParamType | None = None) -> tuple[int, Iterator[Review]]:
        """Lists all reviews and their details"""
        return self._endpoint_lst(Review, event_slug, 'reviews', params=params)

    def room(self, event_slug: str, id: int, *, params: QueryParamType | None = None) -> Room:  # noqa: A002
        """Returns a specific room"""
        return self._endpoint_id(Room, event_slug, 'rooms', id, params=params)

    def rooms(self, event_slug: str, *, params: QueryParamType | None = None) -> tuple[int, Iterator[Room]]:
        """Lists all rooms and their details"""
        return self._endpoint_lst(Room, event_slug, 'rooms', params=params)

    def question(self, event_slug: str, id: int, *, params: QueryParamType | None = None) -> Question:  # noqa: A002
        """Returns a specific question"""
        return self._endpoint_id(Question, event_slug, 'questions', id, params=params)

    def questions(self, event_slug: str, *, params: QueryParamType | None = None) -> tuple[int, Iterator[Question]]:
        """Lists all questions and their details"""
        return self._endpoint_lst(Question, event_slug, 'questions', params=params)

    def answer(self, event_slug: str, id: int, *, params: QueryParamType | None = None) -> Answer:  # noqa: A002
        """Returns a specific answer"""
        return self._endpoint_id(Answer, event_slug, 'answers', id, params=params)

    def answers(self, event_slug: str, *, params: QueryParamType | None = None) -> tuple[int, Iterator[Answer]]:
        """Lists all answers and their details"""
        return self._endpoint_lst(Answer, event_slug, 'answers', params=params)

    def tag(self, event_slug: str, tag: str, *, params: QueryParamType | None = None) -> Tag:
        """Returns a specific tag"""
        return self._endpoint_id(Tag, event_slug, 'tags', tag, params=params)

    def tags(self, event_slug: str, *, params: QueryParamType | None = None) -> tuple[int, Iterator[Tag]]:
        """Lists all tags and their details"""
        return self._endpoint_lst(Tag, event_slug, 'tags', params=params)


def _log_resp(json_resp: list[Any] | dict[Any, Any]):
    """Log everything except of the actual 'results'"""
    if isinstance(json_resp, dict):
        _logger.debug(f"response: {rm_keys('results', json_resp)}")



"""Functionality around the Google's Spreadsheet API

Additional Documentation:
    * [Google GSheet API](https://developers.google.com/sheets/api/quickstart/python)
    * [GSpread](https://docs.gspread.org/)
    * [GSpread-Dataframe](https://gspread-dataframe.readthedocs.io/)
    * [GSpread-Formatting](https://gspread-formatting.readthedocs.io/)
"""

import string
import time
from enum import Enum

import gspread
import numpy as np
import pandas as pd
from gspread.client import APIError
from gspread.spreadsheet import Spreadsheet
from gspread.worksheet import Worksheet
from gspread_dataframe import get_as_dataframe, set_with_dataframe
from gspread_formatting import (
    Color,
    format_cell_range,
    format_cell_ranges,
    get_conditional_format_rules,
    get_default_format,
    set_data_validation_for_cell_range,
)
from gspread_formatting.dataframe import format_with_dataframe
from gspread_formatting.models import cellFormat
from matplotlib.colors import to_rgb
from structlog import get_logger

from pytanis.config import Config, get_cfg

# Color type of matplotlib: https://matplotlib.org/stable/tutorials/colors/colors.html
ColorType = str | tuple[float, float, float] | tuple[float, float, float, float]

__all__ = ['GSheetClient', 'gsheet_rows_for_fmt', 'PermissionDeniedError']

_logger = get_logger()


class Scope(Enum):
    # Allows read-only access to the user's sheets and their properties
    GSHEET_RO = 'https://www.googleapis.com/auth/spreadsheets.readonly'
    # Allows read/write access to the user's sheets and their properties
    GSHEET_RW = 'https://www.googleapis.com/auth/spreadsheets'
    # Allows read-only access to the user's file metadata and file content
    GDRIVE_RO = 'https://www.googleapis.com/auth/drive.readonly'
    # Per-file access to files created or opened by the app
    GDRIVE_FILE = 'https://www.googleapis.com/auth/drive.file'
    # Full, permissive scope to access all of a user's files. Request this scope only when it is strictly necessary
    GDRIVE_RW = 'https://www.googleapis.com/auth/drive'


def gspread_client(scopes: list[Scope], config: Config) -> gspread.client.Client:
    """Creates the GSheet client using our configuration

    Read [GSpread](https://docs.gspread.org/) for usage details
    """
    if (secret_path := config.Google.client_secret_json) is None:
        msg = 'You have to set Google.client_secret_json in your config.toml!'
        raise RuntimeError(msg)
    if (token_path := config.Google.token_json) is None:
        msg = 'You have to set Google.token_json in your config.toml!'
        raise RuntimeError(msg)

    gc = gspread.oauth(
        scopes=[scope.value for scope in scopes],
        credentials_filename=str(secret_path),
        authorized_user_filename=str(token_path),
    )
    return gc


class PermissionDeniedError(Exception):
    """Error for APIError with status PERMISSION_DENIED

    Most likely thrown in cases when the scope is not `GSHEET_RW` or the token needs to be updated accordingly.
    """


class GSheetClient:
    """Google API to easily handle GSheets and other files on GDrive

    By default, only the least permissive scope `GSHEET_RO` in case of `read_only = True` is used.
    """

    def __init__(self, config: Config | None = None, *, read_only: bool = True):
        self._read_only = read_only
        if read_only:
            self._scopes = [Scope.GSHEET_RO]
        else:
            self._scopes = [Scope.GSHEET_RW]
        if config is None:
            config = get_cfg()
        self._config = config
        self.gc = gspread_client(self._scopes, config)  # gspread client for more functionality

    def recreate_token(self):
        """Recreate the current token using the scopes given at initialization"""
        self._config.Google.token_json.unlink(missing_ok=True)
        self.gc = gspread_client(self._scopes, self._config)

    def _wait_for_worksheet(self, spreadsheet_id: str, worksheet_name: str):
        """Wait for the worksheet to come into existence"""
        spreadsheet = self.gc.open_by_key(spreadsheet_id)
        while worksheet_name not in [ws.title for ws in spreadsheet.worksheets()]:
            time.sleep(1)

    def gsheet(
        self, spreadsheet_id: str, worksheet_name: str | None = None, *, create_ws: bool = False
    ) -> Worksheet | Spreadsheet:
        """Retrieve a Google sheet by its id and the name

        Open a Google sheet in your browser and check the URL to retrieve the id, e.g.:
        https://docs.google.com/spreadsheets/d/SPREEDSHEET_ID/edit...

        If the spreadsheet as several worksheets (check the lower bar) then `worksheet_name` can be used to
        specify a specific one.
        """
        spreadsheet = self.gc.open_by_key(spreadsheet_id)
        if worksheet_name is None:
            return spreadsheet
        elif worksheet_name in [ws.title for ws in spreadsheet.worksheets()]:
            return spreadsheet.worksheet(worksheet_name)
        elif create_ws:
            worksheet = spreadsheet.add_worksheet(title=worksheet_name, rows=100, cols=20)
            self._wait_for_worksheet(spreadsheet_id, worksheet_name)
            return worksheet
        else:
            return spreadsheet.worksheet(worksheet_name)  # raises exception

    def _exception_feedback(self, error: APIError):
        if error.response.json()['error']['status'] == 'PERMISSION_DENIED':
            if self._read_only:
                msg = 'For saving `read_only=False` is needed when initializing this client!'
                raise PermissionDeniedError(msg) from error
            else:
                msg = 'Attempt to recreate your current token by calling the method `recreate_token()` first!'
                raise PermissionDeniedError(msg) from error
        else:
            raise error

    def save_df_as_gsheet(
        self,
        df: pd.DataFrame,
        spreadsheet_id: str,
        worksheet_name: str,
        *,
        create_ws: bool = False,
        default_fmt: bool = True,
        **kwargs: str | (bool | int),
    ):
        """Save the given dataframe as worksheet in a spreadsheet

        Make sure that the scope passed gives you write permissions

        Args:
            df: dataframe to save
            spreadsheet_id: id of the Google spreadsheet
            worksheet_name: name of the worksheet within the spreadsheet
            create_ws: create the worksheet if non-existent
            default_fmt: apply default formatter `BasicFormatter`
            **kwargs: extra keyword arguments passed to `set_with_dataframe`
        """
        worksheet = self.gsheet(spreadsheet_id, worksheet_name, create_ws=create_ws)
        # make sure it's really only the dataframe, not some residue
        self.clear_gsheet(spreadsheet_id, worksheet_name)
        params = {'resize': True} | dict(**kwargs)  # set sane defaults
        try:
            set_with_dataframe(worksheet, df, **params)
            if default_fmt:
                format_with_dataframe(worksheet, df)
        except APIError as error:
            self._exception_feedback(error)

    def clear_gsheet(self, spreadsheet_id: str, worksheet_name: str):
        """Clear the worksheet including values, formatting, filtering, etc."""
        worksheet = self.gsheet(spreadsheet_id, worksheet_name, create_ws=False)
        default_fmt = get_default_format(worksheet.spreadsheet)
        wrange = worksheet_range(worksheet)
        try:
            worksheet.clear()
            worksheet.clear_basic_filter()
            format_cell_range(worksheet, wrange, default_fmt)
            rules = get_conditional_format_rules(worksheet)
            rules.clear()
            rules.save()
            set_data_validation_for_cell_range(worksheet, wrange, None)
        except APIError as error:
            self._exception_feedback(error)

    def gsheet_as_df(self, spreadsheet_id: str, worksheet_name: str, **kwargs: str | (bool | int)) -> pd.DataFrame:
        """Returns a worksheet as dataframe"""
        worksheet = self.gsheet(spreadsheet_id, worksheet_name)
        df = get_as_dataframe(worksheet, **kwargs)
        # remove Nan rows & columns as they are exported by default
        df.dropna(how='all', inplace=True, axis=0)
        df.dropna(how='all', inplace=True, axis=1)
        return df


def gsheet_col(idx: int) -> str:
    """Convert a column index to Google Sheet range notation, e.g. A, BE, etc."""
    idx += 1
    chars = []
    while idx:
        chars.append(string.ascii_uppercase[(idx % 26) - 1])
        idx //= 27
    return ''.join(chars[::-1])


def gsheet_rows_for_fmt(mask: pd.Series, n_cols: int) -> list[str]:
    """Get the Google Sheet row range specifications for formatting"""
    rows = pd.Series(np.argwhere(mask.to_numpy()).reshape(-1) + 2)  # +2 since 1-index and header
    last_col = gsheet_col(n_cols - 1)  # last index
    rows = rows.map(lambda x: f'A{x}:{last_col}{x}')
    return rows.to_list()


def worksheet_range(worksheet: Worksheet) -> str:
    """Returns a range encompassing the whole worksheet"""
    last_row = worksheet.row_count
    last_col = gsheet_col(worksheet.col_count)
    return f'A1:{last_col}{last_row}'


def mark_rows(worksheet, mask: pd.Series, color: ColorType):
    """Mark rows specified by a mask (condition) with a given color

    Color can be a tuple of RGB values or a Matplotlib string specification:
    https://matplotlib.org/stable/gallery/color/named_colors.html#css-colors
    """
    rows = gsheet_rows_for_fmt(mask, worksheet.col_count)
    fmt = cellFormat(backgroundColor=Color(*to_rgb(color)))
    if rows:
        format_cell_ranges(worksheet, [(rng, fmt) for rng in rows])



"""Handling the configuration"""

import os
from pathlib import Path

import tomli
from pydantic import BaseModel, FilePath, validator

PYTANIS_ENV: str = 'PYTANIS_CONFIG'
"""Name of the environment variable to look up the path for the config"""
PYTANIS_CFG_PATH: str = '.pytanis/config.toml'
"""Path within $HOME to the configuration file of Pytanis"""


class Google(BaseModel):
    """Configuration related to the Google API"""

    client_secret_json: Path | None = None
    token_json: Path | None = None


class HelpDesk(BaseModel):
    """Configuration related to the HelpDesk API"""

    account: str | None = None
    entity_id: str | None = None
    token: str | None = None


class Pretalx(BaseModel):
    """Configuration related to the Pretalx API"""

    api_token: str | None = None


class Config(BaseModel):
    """Main configuration object"""

    cfg_path: FilePath

    Pretalx: Pretalx
    Google: Google
    HelpDesk: HelpDesk

    # TODO[pydantic]: We couldn't refactor the `validator`, please replace it by `field_validator` manually.
    # Check https://docs.pydantic.dev/dev-v2/migration/#changes-to-validators for more information.
    @validator('Google')
    @classmethod
    def convert_json_path(cls, v, values):
        def make_rel_path_abs(entry):
            if entry is not None and not entry.is_absolute():
                entry = values['cfg_path'].parent / entry
            return entry

        v.client_secret_json = make_rel_path_abs(v.client_secret_json)
        v.token_json = make_rel_path_abs(v.token_json)

        return v


def get_cfg_file() -> Path:
    """Determines the path of the config file"""
    path_str = os.environ.get(PYTANIS_ENV, None)
    path = Path.home() / Path(PYTANIS_CFG_PATH) if path_str is None else Path(path_str)
    return path


def get_cfg() -> Config:
    """Returns the configuration as an object"""
    cfg_path = get_cfg_file()
    with open(cfg_path, 'rb') as fh:
        cfg_dict = tomli.load(fh)
    # add config path to later resolve relative paths of config values
    cfg_dict['cfg_path'] = cfg_path
    return Config.parse_obj(cfg_dict)



"""Utilities related to Pretalx"""

from collections.abc import Iterable

import pandas as pd

from pytanis.pretalx.types import Review, Speaker, Submission


class Col:
    """Convention of Pretalx column names for the functions below."""

    submission = 'Submission'
    submission_type = 'Submission type'
    submission_type_id = 'Submission type id'
    title = 'Title'
    duration = 'Duration'
    public = 'Public'
    track = 'Track'
    comment = 'Comment'
    created = 'Created'
    state = 'State'
    pending_state = 'Pending state'

    speaker_name = 'Speaker name'
    speaker_code = 'Speaker code'
    pretalx_user = 'Pretalx user'
    biography = 'Biography'
    affiliation = 'Affiliation'
    email = 'Email'
    availability = 'Availability'
    availability_comment = 'Availability Comment'

    nreviews = '#Reviews'
    review_score = 'Review Score'


def subs_as_df(
    subs: Iterable[Submission], *, with_questions: bool = False, question_prefix: str = 'Q: '
) -> pd.DataFrame:
    """Convert submissions into a dataframe

    Make sure to have `params={"questions": "all"}` for the PretalxAPI if `with_questions` is True.
    """
    rows = []
    for sub in subs:
        row = {
            Col.submission: sub.code,
            Col.title: sub.title,
            Col.track: sub.track.en if sub.track else None,
            Col.speaker_code: [speaker.code for speaker in sub.speakers],
            Col.speaker_name: [speaker.name for speaker in sub.speakers],
            Col.duration: sub.duration,
            Col.submission_type: sub.submission_type.en,
            Col.submission_type_id: sub.submission_type_id,
            Col.state: sub.state.value,
            Col.pending_state: None if sub.pending_state is None else sub.pending_state.value,
            Col.created: sub.created,
        }
        if with_questions and sub.answers is not None:
            for answer in sub.answers:
                row[f'{question_prefix}{answer.question.question.en}'] = answer.answer
        rows.append(row)
    return pd.DataFrame(rows)


def speakers_as_df(
    speakers: Iterable[Speaker], *, with_questions: bool = False, question_prefix: str = 'Q: '
) -> pd.DataFrame:
    """Convert speakers into a dataframe

    Make sure to have `params={"questions": "all"}` for the PretalxAPI if `with_questions` is True.
    """
    rows = []
    for speaker in speakers:
        row = {
            Col.speaker_code: speaker.code,
            Col.speaker_name: speaker.name,
            Col.email: speaker.email,
            Col.biography: speaker.biography,
            Col.submission: speaker.submissions,
        }
        if with_questions and speaker.answers is not None:
            for answer in speaker.answers:
                # The API returns also questions that are 'per proposal/submission', we get these using the
                # submission endpoint and don't want them here due to ambiguity if several submission were made.
                if answer.person is not None:
                    row[f'{question_prefix}{answer.question.question.en}'] = answer.answer
        rows.append(row)
    return pd.DataFrame(rows)


def reviews_as_df(reviews: Iterable[Review]) -> pd.DataFrame:
    """Convert the reviews to a dataframe"""
    df = pd.DataFrame([review.model_dump() for review in reviews])
    # make first letter of column upper-case in accordance with our convention
    df.rename(columns={col: col.title() for col in df.columns}, inplace=True)
    # user is the speaker name to use for joining
    df.rename(columns={'User': Col.pretalx_user, 'Score': Col.review_score}, inplace=True)

    return df



"""Functionality around the Pretalx API"""

from pytanis.pretalx.client import PretalxClient
from pytanis.pretalx.utils import reviews_as_df, speakers_as_df, subs_as_df

__all__ = ['PretalxClient', 'subs_as_df', 'speakers_as_df', 'reviews_as_df']



"""Functionality around mailing

ToDo:
    * add logging where appropriate
    * Find out why `extra=Extra.allow` causes mypy to fail. Seems like a bug in pydantic.
    * Sending mails is quite slow, so using `tqdm` to show feedback to the current progress would be nice
"""

from collections.abc import Callable

from pydantic import BaseModel, validator
from structlog import get_logger
from tqdm.auto import tqdm

from pytanis.helpdesk.client import HelpDeskClient
from pytanis.helpdesk.types import Assignment, Id, Message, NewTicket, Requester, Ticket

_logger = get_logger()


class MetaData(BaseModel, extra='allow'):  # type: ignore
    """Additional, arbitrary metadata provided by the user like for template filling"""


class Recipient(BaseModel):
    """Details about the recipient

    Use the `data` field to store additional information
    """

    name: str
    email: str
    address_as: str | None = None  # could be the first name
    data: MetaData | None = None

    # TODO[pydantic]: We couldn't refactor the `validator`, please replace it by `field_validator` manually.
    # Check https://docs.pydantic.dev/dev-v2/migration/#changes-to-validators for more information.
    @validator('address_as')
    @classmethod
    def fill_with_name(cls, v, values):
        if v is None:
            v = values['name']
        return v


class Mail(BaseModel):
    """Mail template

    Use the `data` field to store additional information

    You can use the typical [Format String Syntax] and the objects `recipient` and `mail`
    to access metadata to complement the template, e.g.:

    ```
    Hello {recipient.address_as},

    We hope it's ok to address you your first name rather than using your full name being {recipient.name}.
    Have you read the email's subject '{mail.subject}'? How is your work right now at {recipient.data.company}?

    Cheers!
    ```

    [Format String Syntax]: https://docs.python.org/3/library/string.html#formatstrings
    """

    subject: str
    team_id: str
    agent_id: str
    text: str
    status: str = 'solved'  # ToDo: Reconsider this!
    recipients: list[Recipient]
    data: MetaData | None = None


class MailClient:
    """Mail client for mass mails over HelpDesk"""

    def __init__(self, helpdesk_client: HelpDeskClient | None = None):
        if helpdesk_client is None:
            helpdesk_client = HelpDeskClient()
        self._helpdesk_client = helpdesk_client
        self.dry_run: Callable[[NewTicket], None] = self.print_new_ticket

    @staticmethod
    def _create_ticket(mail: Mail, recipient: Recipient) -> NewTicket:
        message = Message(text=mail.text)
        requester = Requester(name=recipient.name, email=recipient.email)
        team_id, agent_id = Id(ID=mail.team_id), Id(ID=mail.agent_id)
        assignment = Assignment(team=team_id, agent=agent_id)
        ticket = NewTicket(
            requester=requester,
            message=message,
            subject=mail.subject,
            assignment=assignment,
            status=mail.status,
            teamIDs=[mail.team_id],
        )
        return ticket

    @staticmethod
    def print_new_ticket(ticket: NewTicket):
        """Default action in a dry-run. Mainly for making sure you sent what you mean!

        Overwrite it by assigning to self.dry_run another function

        ToDo: Make this function nice, maybe use the `rich` library even
        """
        print('#' * 40)  # noqa: T201
        print(f'Recipient: {ticket.requester.name} <{ticket.requester.email}>')  # noqa: T201
        print(f'Subject: {ticket.subject}')  # noqa: T201
        print(f'{ticket.message.text}')  # noqa: T201

    def send(
        self, mail: Mail, *, dry_run: bool = True
    ) -> tuple[list[tuple[Recipient, Ticket | None]], list[tuple[Recipient, Exception]]]:
        """Send a mail to all recipients using HelpDesk"""
        errors = []
        tickets = []
        for recipient in tqdm(mail.recipients):
            recip_mail = mail.model_copy()
            try:
                recip_mail.subject = mail.subject.format(recipient=recipient, mail=mail)
                # be aware here that the body might reference to subject line, so it must be filled already
                recip_mail.text = recip_mail.text.format(recipient=recipient, mail=recip_mail)
                ticket = self._create_ticket(recip_mail, recipient)
                if dry_run:
                    self.print_new_ticket(ticket)
                    resp_ticket = None
                else:
                    resp = self._helpdesk_client.create_ticket(ticket)
                    resp_ticket = Ticket.model_validate(resp)
            except Exception as e:
                errors.append((recipient, e))
            else:
                tickets.append((recipient, resp_ticket))
        return tickets, errors



"""Functionality around the HelpDesk / LiveChat API"""

from pytanis.helpdesk.client import HelpDeskClient
from pytanis.helpdesk.mail import Mail, MailClient, Recipient

__all__ = ['HelpDeskClient', 'Mail', 'MailClient', 'Recipient']



"""Tools related to assigning proposals to reviewers

In Pretalx assignments can be done in two directions:

 1. Assign proposals to reviewers
 2. Assign reviewers to proposals

We will always assume direction 1. in this file when we talk about an assignment.
So in Operation Research-speak, resources get assigned tasks, not the other way around.
The time needed for the task of reviewing a proposal is quite homogeneous while the number of reviews a single
reviewer may highly vary. Also, we will rather use the name submission instead of proposal as this also reflects
the naming of the Pretalx API.

We follow the convention over configuration principle here and thus check out the `Col` class for the naming of
columns.
"""

import json
from pathlib import Path

import pandas as pd

from pytanis.pretalx.utils import Col as PretalxCol


class Col(PretalxCol):
    """Additional conventions used for reviews"""

    address_as = 'Address as'
    track_prefs = 'Track Preferences'
    committee_contact = 'Committee Contact'
    committee_member = 'Committee Member'
    all_proposals = 'All Proposals'

    pretalx_activated = 'Pretalx activated'

    curr_assignments = 'Current Assignments'
    nassignments = '#Assignments'

    target_nreviews = 'Target #Reviews'
    rem_nreviews = 'Remaining #Reviews'
    done_nreviews = 'Done #Reviews'

    nvotes = '#Votes'
    vote_score = 'Vote Score'


def read_assignment_as_df(file_path: Path) -> pd.DataFrame:
    """Reads an assignment and returns a dataframe"""
    with open(file_path, encoding='utf8') as fh:
        curr_assign = json.load(fh)
    df = pd.DataFrame({k: [v] for k, v in curr_assign.items()})
    df = df.T.rename_axis(index=Col.email).rename(columns={0: Col.curr_assignments}).reset_index()
    return df


def save_assignments_as_json(df: pd.DataFrame, file_path: Path | str):
    """Save the dataframe as proposal assignment JSON file"""
    file_path = Path(file_path)
    df = df.loc[:, [Col.email, Col.curr_assignments]]
    json_dct = json.loads(df.set_index(Col.email).to_json())[Col.curr_assignments]
    # prettify the json string for human-edit-ability if reviewers need to be dropped later
    json_str = json.dumps(json_dct).replace('{', '{\n').replace('], ', '],\n').replace(']}', ']\n}')
    with open(file_path, 'w', encoding='utf8') as fh:
        fh.write(json_str)



"""Client for the HelpDesk / LiveChat API

Documentation: https://api.helpdesk.com/docs

ToDo:
    * Transfer more functionality from https://github.com/PYCONDE/py_helpdesk_com
"""

from typing import Any, TypeAlias

import httpx
from httpx import URL, Response
from httpx_auth import Basic
from structlog import get_logger

from pytanis.config import Config, get_cfg
from pytanis.helpdesk.types import Agent, NewTicket, Team
from pytanis.utils import throttle

_logger = get_logger()


JSONObj: TypeAlias = dict[str, Any]
"""Type of a JSON object (without recursion)"""
JSONLst: TypeAlias = list[JSONObj]
"""Type of a JSON list of JSON objects"""
JSON: TypeAlias = JSONObj | JSONLst
"""Type of the JSON response as returned by the HelpDesk / LiveChat API"""


class HelpDeskClient:
    def __init__(self, config: Config | None = None):
        if config is None:
            config = get_cfg()
        self._config = config
        # Important: Always use a custom User-Agent, never a generic one.
        # Generic User-Agents are filtered by helpdesk to reduce spam.
        self._headers = {'User-Agent': 'Pytanis'}

        self._get_throttled = self._get
        self._post_throttled = self._post
        self.set_throttling(1, 1)  # we are nice by default

    def set_throttling(self, calls: int, seconds: int):
        """Throttle the number of calls per seconds to the Pretalx API"""
        _logger.debug('throttling', calls=calls, seconds=seconds)
        self._get_throttled = throttle(calls, seconds)(self._get)
        self._post_throttled = throttle(calls, seconds)(self._post)

    def _get(self, endpoint: str, params: dict[str, str] | None = None) -> Response:
        """Retrieve data via raw GET request"""
        auth = Basic(self._config.HelpDesk.account, self._config.HelpDesk.token)
        url = URL('https://api.helpdesk.com/v1/').join(endpoint)
        _logger.debug(f'GET: {url.copy_merge_params(params)}')
        return httpx.get(url, auth=auth, params=params, headers=self._headers)

    def get(self, endpoint: str, params: dict[str, str] | None = None) -> JSON:
        """Retrieve data via throttled GET request and return the JSON"""
        resp = self._get_throttled(endpoint, params)
        resp.raise_for_status()
        return resp.json()

    def _post(self, endpoint: str, data: dict[str, Any], params: dict[str, str] | None = None) -> Response:
        """Sent data via raw POST request"""
        auth = Basic(self._config.HelpDesk.account, self._config.HelpDesk.token)
        url = URL('https://api.helpdesk.com/v1/').join(endpoint)
        _logger.debug(f'POST: {url.copy_merge_params(params)}')
        return httpx.post(url, auth=auth, params=params, json=data, headers=self._headers)

    def post(self, endpoint: str, data: dict[str, Any], params: dict[str, str] | None = None) -> JSON:
        resp = self._post_throttled(endpoint, data, params)
        resp.raise_for_status()
        return resp.json()

    def list_agents(self) -> list[Agent]:
        agents = self.get('agents')
        if not isinstance(agents, list):
            msg = 'Received JSON is not a list object'
            raise ValueError(msg)
        return [Agent.model_validate(dct) for dct in agents]

    def list_teams(self) -> list[Team]:
        teams = self.get('teams')
        if not isinstance(teams, list):
            msg = 'Received JSON is not a list object'
            raise ValueError(msg)
        return [Team.model_validate(dct) for dct in teams]

    def create_ticket(self, ticket: NewTicket):
        return self.post('tickets', data=ticket.model_dump())
